# Copyright (c) 2020-2021 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
from unittest import mock

from oslotest.base import BaseTestCase

from sushy_oem_idrac.resources.manager import job_collection


class DellJobCollectionTestCase(BaseTestCase):

    def setUp(self):
        super(DellJobCollectionTestCase, self).setUp()
        self.conn = mock.Mock()
        with open('sushy_oem_idrac/tests/unit/json_samples/'
                  'job_collection_expanded.json') as f:
            mock_response = self.conn.get.return_value
            mock_response.json.return_value = json.load(f)
            mock_response.status_code = 200

        mock_response = self.conn.post.return_value
        mock_response.status_code = 202
        mock_response.headers.get.return_value = '1'
        self.job_collection = job_collection.DellJobCollection(
            self.conn, '/redfish/v1/Managers/iDRAC.Embedded.1/Jobs')

    def test_get_unfinished_jobs(self):
        expected_unfinished_jobs = ['RID_878460711202']
        actual_unfinished_jobs = self.job_collection.get_unfinished_jobs()
        target_uri = ('/redfish/v1/Managers/iDRAC.Embedded.1'
                      '/Jobs?$expand=.($levels=1)')
        self.conn.get.assert_called_with(target_uri)
        self.assertEqual(expected_unfinished_jobs, actual_unfinished_jobs)



# Copyright (c) 2020-2021 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from sushy import exceptions
from sushy.resources import base

from sushy_oem_idrac.resources.manager import constants as mgr_cons
from sushy_oem_idrac.resources.manager import mappings as mgr_maps

LOG = logging.getLogger(__name__)


class ForceActionField(base.CompositeField):
    target_uri = base.Field('target', required=True)
    allowed_values = base.Field('Force@Redfish.AllowableValues',
                                adapter=list)


class ActionsField(base.CompositeField):
    reset_idrac = ForceActionField('#DelliDRACCardService.iDRACReset')


class DelliDRACCardService(base.ResourceBase):

    _actions = ActionsField('Actions')
    identity = base.Field('Id', required=True)

    def __init__(self, connector, identity, redfish_version=None,
                 registries=None):
        """A class representing a DelliDRACCardService.

        :param connector: A Connector instance
        :param identity: The identity of the DelliDRACCardService resource
        :param redfish_version: The version of Redfish. Used to construct
            the object according to schema of the given version.
        :param registries: Dict of Redfish Message Registry objects to be
            used in any resource that needs registries to parse messages.
        """
        super(DelliDRACCardService, self).__init__(
            connector, identity, redfish_version, registries)

    def get_allowed_reset_idrac_values(self):
        """Get the allowed values for resetting the idrac.

        :returns: A set of allowed values.
        """
        reset_action = self._actions.reset_idrac

        if not reset_action.allowed_values:
            LOG.warning('Could not figure out the allowed values for the '
                        'reset idrac action for %s', self.identity)
            return set(mgr_maps.RESET_IDRAC_VALUE_MAP_REV)

        return set([mgr_maps.RESET_IDRAC_VALUE_MAP[value] for value in
                    set(mgr_maps.RESET_IDRAC_VALUE_MAP).
                   intersection(reset_action.allowed_values)])

    def reset_idrac(self):
        """Reset the iDRAC.

        """
        reset_type = mgr_cons.RESET_IDRAC_GRACEFUL_RESTART
        valid_resets = self.get_allowed_reset_idrac_values()
        if reset_type not in valid_resets:
            raise exceptions.InvalidParameterValueError(
                parameter='value', value=reset_type, valid_values=valid_resets)
        reset_type = mgr_maps.RESET_IDRAC_VALUE_MAP_REV[reset_type]
        target_uri = self._actions.reset_idrac.target_uri
        payload = {"Force": reset_type}
        LOG.debug('Resetting the iDRAC %s ...', self.identity)
        self._conn.post(target_uri, data=payload)
        LOG.info('The iDRAC %s is being reset', self.identity)



# Copyright (c) 2020-2021 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from sushy import utils

from sushy_oem_idrac.resources.manager import constants as mgr_cons

EXPORT_CONFIG_VALUE_MAP = {
    'ALL': mgr_cons.EXPORT_TARGET_ALL,
    'BIOS': mgr_cons.EXPORT_TARGET_BIOS,
    'IDRAC': mgr_cons.EXPORT_TARGET_IDRAC,
    'NIC': mgr_cons.EXPORT_TARGET_NIC,
    'RAID': mgr_cons.EXPORT_TARGET_RAID
}

EXPORT_CONFIG_VALUE_MAP_REV = utils.revert_dictionary(EXPORT_CONFIG_VALUE_MAP)

RESET_IDRAC_VALUE_MAP = {
    'Graceful': mgr_cons.RESET_IDRAC_GRACEFUL_RESTART,
    'Force': mgr_cons.RESET_IDRAC_FORCE_RESTART,
}

RESET_IDRAC_VALUE_MAP_REV = utils.revert_dictionary(RESET_IDRAC_VALUE_MAP)

IMPORT_SHUTDOWN_VALUE_MAP = {
    'Graceful': mgr_cons.IMPORT_SHUTDOWN_GRACEFUL,
    'Forced': mgr_cons.IMPORT_SHUTDOWN_FORCED,
    'NoReboot': mgr_cons.IMPORT_SHUTDOWN_NO_REBOOT
}

IMPORT_SHUTDOWN_VALUE_MAP_REV =\
    utils.revert_dictionary(IMPORT_SHUTDOWN_VALUE_MAP)

EXPORT_USE_VALUE_MAP = {
    'Default': mgr_cons.EXPORT_USE_DEFAULT,
    'Clone': mgr_cons.EXPORT_USE_CLONE,
    'Replace': mgr_cons.EXPORT_USE_REPLACE
}

EXPORT_USE_VALUE_MAP_REV = utils.revert_dictionary(EXPORT_USE_VALUE_MAP)

INCLUDE_EXPORT_VALUE_MAP = {
    'Default': mgr_cons.INCLUDE_EXPORT_DEFAULT,
    'IncludeReadOnly': mgr_cons.INCLUDE_EXPORT_READ_ONLY,
    'IncludePasswordHashValues':
        mgr_cons.INCLUDE_EXPORT_PASSWORD_HASHES,
    'IncludeReadOnly,IncludePasswordHashValues':
        mgr_cons.INCLUDE_EXPORT_READ_ONLY_PASSWORD_HASHES
}

INCLUDE_EXPORT_VALUE_MAP_REV =\
    utils.revert_dictionary(INCLUDE_EXPORT_VALUE_MAP)



# Copyright (c) 2020-2021 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# export system config action constants

EXPORT_TARGET_ALL = 'all'
"""Export entire system configuration"""

EXPORT_TARGET_BIOS = 'BIOS'
"""Export BIOS related configuration"""

EXPORT_TARGET_IDRAC = 'iDRAC'
"""Export IDRAC related configuration"""

EXPORT_TARGET_NIC = 'NIC'
"""Export NIC related configuration"""

EXPORT_TARGET_RAID = 'RAID'
"""Export RAID related configuration"""

# iDRAC Reset action constants


RESET_IDRAC_GRACEFUL_RESTART = 'graceful restart'
"""Perform a graceful shutdown followed by a restart of the system"""

RESET_IDRAC_FORCE_RESTART = 'force restart'
"""Perform an immediate (non-graceful) shutdown, followed by a restart"""

# ImportSystemConfiguration ShutdownType values
IMPORT_SHUTDOWN_GRACEFUL = 'graceful shutdown'
"""Graceful shutdown for Import System Configuration

Will wait for the host up to 5 minutes to shut down before timing out. The
operating system can potentially deny or ignore the graceful shutdown request.
"""

IMPORT_SHUTDOWN_FORCED = 'forced shutdown'
"""Forced shutdown for Import System Configuration

The host server will be powered off immediately. Should be used when it is safe
to power down the host.
"""

IMPORT_SHUTDOWN_NO_REBOOT = 'no shutdown'
"""No reboot for Import System Configuration

No shutdown performed. Explicit reboot is necessary to apply changes.
"""

# ExportUse in ExportSystemConfiguration
EXPORT_USE_DEFAULT = 'Default'
"""Default export type

Leaves some attributes commented out and requires user to enable them before
they can be applied during import.
"""

EXPORT_USE_CLONE = 'Clone'
"""Clone export type suitable for cloning a 'golden' configuration.

Compared to Default export type, more attributes are enabled and
storage settings adjusted to aid in cloning process.
"""

EXPORT_USE_REPLACE = 'Replace'
"""Replace export type suited for retiring or replacing complete configuration.

Compared to Clone export type, most attributes are enabled and storage settings
adjusted to aid in the replace process.
"""

# IncludeInExport in ExportSystemConfiguration
INCLUDE_EXPORT_DEFAULT = 'Default'
"""Default for what to include in export.

Does not include read-only attributes, and depending on Export Use, passwords
are marked as ****** (for Default) or are set to default password values (for
Clone and Replace).
"""

INCLUDE_EXPORT_READ_ONLY = 'Include read only attributes'
"""Includes read-only attributes.

In addition to values included by Default option, this also includes read-only
attributes that cannot be changed via Import and are provided for informational
purposes only.
"""

INCLUDE_EXPORT_PASSWORD_HASHES = 'Include password hash values'
"""Include password hashes.

When using Clone or Replace, include password hashes, instead of default
password. Can be used to replicate passwords across systems.
"""

INCLUDE_EXPORT_READ_ONLY_PASSWORD_HASHES = ('Include read only attributes and '
                                            'password hash values')
"""Includes both read-only attributes and password hashes.

INCLUDE_EXPORT_READ_ONLY and INCLUDE_EXPORT_PASSWORD_HASHES combined
"""



# Copyright (c) 2020-2021 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import subprocess
import time
from urllib.parse import urlparse

import sushy
from sushy.resources import base
from sushy.resources import common
from sushy.resources.oem import base as oem_base
from sushy.taskmonitor import TaskMonitor
from sushy import utils as sushy_utils

from sushy_oem_idrac import asynchronous
from sushy_oem_idrac import constants
from sushy_oem_idrac.resources.manager import constants as mgr_cons
from sushy_oem_idrac.resources.manager import idrac_card_service
from sushy_oem_idrac.resources.manager import job_collection
from sushy_oem_idrac.resources.manager import job_service
from sushy_oem_idrac.resources.manager import lifecycle_service
from sushy_oem_idrac.resources.manager import mappings as mgr_maps
from sushy_oem_idrac import utils

LOG = logging.getLogger(__name__)

# System Configuration Tag Constant
_SYSTEM_CONFIG_TAG = "SystemConfiguration"

# Response Code Constant
_RESPONSE_OK_CODE = 200


class SharedParameters(base.CompositeField):
    allowed_target_values = base.Field('Target@Redfish.AllowableValues')


class ExportActionField(common.ActionField):
    shared_parameters = SharedParameters('ShareParameters')
    allowed_export_use_values = base.Field(
        'ExportUse@Redfish.AllowableValues', adapter=list)
    allowed_include_in_export_values = base.Field(
        'IncludeInExport@Redfish.AllowableValues', adapter=list)


class ImportActionField(common.ActionField):
    allowed_shutdown_type_values = base.Field(
        'ShutdownType@Redfish.AllowableValues', adapter=list)


class DellManagerActionsField(base.CompositeField):
    import_system_configuration = ImportActionField(
        lambda key, **kwargs: key.endswith(
            '#OemManager.ImportSystemConfiguration'))

    export_system_configuration = ExportActionField(
        lambda key, **kwargs: key.endswith(
            '#OemManager.ExportSystemConfiguration'))


class DellManagerExtension(oem_base.OEMResourceBase):

    _actions = DellManagerActionsField('Actions')

    ACTION_DATA = {
        'ShareParameters': {
            'Target': 'ALL'
        },
        'ImportBuffer': None
    }

    # NOTE(etingof): iDRAC job would fail if this XML has
    # insignificant whitespaces

    IDRAC_CONFIG_CD = """\
<SystemConfiguration>\
<Component FQDD="%s">\
<Attribute Name="ServerBoot.1#BootOnce">\
%s\
</Attribute>\
<Attribute Name="ServerBoot.1#FirstBootDevice">\
VCD-DVD\
</Attribute>\
</Component>\
</SystemConfiguration>\
"""

    IDRAC_CONFIG_FLOPPY = """\
<SystemConfiguration>\
<Component FQDD="%s">\
<Attribute Name="ServerBoot.1#BootOnce">\
%s\
</Attribute>\
<Attribute Name="ServerBoot.1#FirstBootDevice">\
VFDD\
</Attribute>\
</Component>\
</SystemConfiguration>\
"""

    IDRAC_MEDIA_TYPES = {
        sushy.VIRTUAL_MEDIA_FLOPPY: IDRAC_CONFIG_FLOPPY,
        sushy.VIRTUAL_MEDIA_CD: IDRAC_CONFIG_CD
    }

    RETRY_COUNT = 35
    RETRY_DELAY = 15

    _IDRAC_IS_READY_RETRIES = 96
    _IDRAC_IS_READY_RETRY_DELAY_SEC = 10

    @property
    def import_system_configuration_uri(self):
        return self._actions.import_system_configuration.target_uri

    @property
    def export_system_configuration_uri(self):
        return self._actions.export_system_configuration.target_uri

    @property
    @sushy_utils.cache_it
    def idrac_card_service(self):
        """Property to reference `DelliDRACCardService` instance of this manager.

        """
        path = sushy_utils.get_sub_resource_path_by(
            self, ["Links", "Oem", "Dell", "DelliDRACCardService"],
            is_collection=False)

        return idrac_card_service.DelliDRACCardService(
            self._conn, path, self.redfish_version, self.registries)

    @property
    @sushy_utils.cache_it
    def lifecycle_service(self):
        """Property to reference `DellLCService` instance of this manager.

        """
        path = sushy_utils.get_sub_resource_path_by(
            self, ["Links", "Oem", "Dell", "DellLCService"],
            is_collection=False)

        return lifecycle_service.DellLCService(
            self._conn, path, self.redfish_version, self.registries)

    @property
    @sushy_utils.cache_it
    def job_service(self):
        """Property to reference `DellJobService` instance of this manager.

        """
        path = sushy_utils.get_sub_resource_path_by(
            self, ["Links", "Oem", "Dell", "DellJobService"],
            is_collection=False)

        return job_service.DellJobService(
            self._conn, path, self.redfish_version, self.registries)

    @property
    @sushy_utils.cache_it
    def job_collection(self):
        """Property to reference `DellJobService` instance of this manager.

        """
        path = sushy_utils.get_sub_resource_path_by(
            self, ["Links", "Oem", "Dell", "Jobs"], is_collection=False)

        return job_collection.DellJobCollection(
            self._conn, path, self.redfish_version, self.registries)

    def set_virtual_boot_device(self, device, persistent=False,
                                manager=None, system=None):
        """Set boot device for a node.

        Dell iDRAC Redfish implementation does not support setting
        boot device to virtual media via standard Redfish means.
        However, this still can be done via an OEM extension.

        :param device: Boot device. Values are vendor-specific.
        :param persistent: Whether to set next-boot, or make the change
            permanent. Default: False.
        :param manager: Manager of OEM extension. Optional.
        :param system: System of OEM extension. Optional.
        :raises: InvalidParameterValue if Dell OEM extension can't
            be used.
        :raises: ExtensionError on failure to perform requested
            operation.
        """
        try:
            idrac_media = self.IDRAC_MEDIA_TYPES[device]

        except KeyError:
            raise sushy.exceptions.InvalidParameterValue(
                error='Unknown or unsupported device %s' % device)

        idrac_media = idrac_media % (
            manager.identity if manager else self._parent_resource.identity,
            'Disabled' if persistent else 'Enabled')

        action_data = dict(self.ACTION_DATA, ImportBuffer=idrac_media)

        # TODO(etingof): figure out if on-time or persistent boot can at
        # all be implemented via this OEM call

        attempts = self.RETRY_COUNT
        rebooted = False

        while True:
            try:
                response = asynchronous.http_call(
                    self._conn, 'post',
                    self.import_system_configuration_uri,
                    data=action_data,
                    sushy_task_poll_period=1)

                LOG.info("Set boot device to %(device)s via "
                         "Dell OEM magic spell (%(retries)d "
                         "retries)", {'device': device,
                                      'retries': self.RETRY_COUNT - attempts})

                return response

            except (sushy.exceptions.ServerSideError,
                    sushy.exceptions.BadRequestError) as exc:
                LOG.warning(
                    'Dell OEM set boot device failed (attempts left '
                    '%d): %s', attempts, exc)

                errors = exc.body and exc.body.get(
                    '@Message.ExtendedInfo') or []

                for error in errors:
                    message_id = error.get('MessageId')

                    LOG.warning('iDRAC error: %s',
                                error.get('Message', 'Unknown error'))

                    if constants.IDRAC_CONFIG_PENDING in message_id:
                        if not rebooted:
                            LOG.warning(
                                'Let\'s try to turn it off and on again... '
                                'This may consume one-time boot settings if '
                                'set previously!')
                            utils.reboot_system(system)
                            rebooted = True
                            break

                    elif constants.IDRAC_JOB_RUNNING in message_id:
                        pass

                else:
                    time.sleep(self.RETRY_DELAY)

                if not attempts:
                    LOG.error('Too many (%d) retries, bailing '
                              'out.', self.RETRY_COUNT)
                    raise

                attempts -= 1

    def get_allowed_export_target_values(self):
        """Get the allowed targets of export system configuration.

        :returns: A set of allowed values.
        """
        export_action = self._actions.export_system_configuration
        allowed_values = export_action.shared_parameters.allowed_target_values

        if not allowed_values:
            LOG.warning('Could not figure out the allowed values for the '
                        'target of export system configuration at %s',
                        self.path)
            return set(mgr_maps.EXPORT_CONFIG_VALUE_MAP_REV)

        return set([mgr_maps.EXPORT_CONFIG_VALUE_MAP[value] for value in
                    set(mgr_maps.EXPORT_CONFIG_VALUE_MAP).
                    intersection(allowed_values)])

    def get_allowed_export_use_values(self):
        """Get allowed export use values of export system configuration.

        :returns: A set of allowed export use values.
        """
        export_action = self._actions.export_system_configuration
        allowed_values = export_action.allowed_export_use_values

        if not allowed_values:
            LOG.warning('Could not figure out the allowed values for the '
                        'export use of export system configuration at %s',
                        self.path)
            return set(mgr_maps.EXPORT_USE_VALUE_MAP_REV)

        return set([mgr_maps.EXPORT_USE_VALUE_MAP[value] for value in
                    set(mgr_maps.EXPORT_USE_VALUE_MAP).
                    intersection(allowed_values)])

    def get_allowed_include_in_export_values(self):
        """Get allowed include in export values of export system configuration.

        :returns: A set of allowed include in export values.
        """
        export_action = self._actions.export_system_configuration
        allowed_values = export_action.allowed_include_in_export_values

        if not allowed_values:
            LOG.warning('Could not figure out the allowed values for the '
                        'include in export of export system configuration at '
                        '%s', self.path)
            return set(mgr_maps.INCLUDE_EXPORT_VALUE_MAP_REV)

        return set([mgr_maps.INCLUDE_EXPORT_VALUE_MAP[value] for value
                   in set(mgr_maps.INCLUDE_EXPORT_VALUE_MAP).
                   intersection(allowed_values)])

    def _export_system_configuration(
        self, target, export_use=mgr_cons.EXPORT_USE_DEFAULT,
        include_in_export=mgr_cons.INCLUDE_EXPORT_DEFAULT):
        """Export system configuration.

        It exports system configuration for specified target like NIC, BIOS,
        RAID and allows to configure purpose for export and what to include.

        :param target: Component of the system to export the
            configuration from. Can be the entire system.
            Valid values can be gotten from
            `get_allowed_export_system_config_values`.
        :param export_use: Export use. Optional, defaults to "Default".
            Valid values can be gotten from `get_allowed_export_use_values`.
        :param include_in_export: What to include in export. Optional. Defaults
            to "Default". Valid values can be gotten from
            `get_allowed_include_in_export_values`.
        :returns: Response object containing configuration details.
        :raises: InvalidParameterValueError on invalid target.
        :raises: ExtensionError on failure to perform requested
            operation
        """
        valid_allowed_targets = self.get_allowed_export_target_values()
        if target not in valid_allowed_targets:
            raise sushy.exceptions.InvalidParameterValueError(
                parameter='target', value=target,
                valid_values=valid_allowed_targets)

        allowed_export_use = self.get_allowed_export_use_values()
        if export_use not in allowed_export_use:
            raise sushy.exceptions.InvalidParameterValueError(
                parameter='export_use', value=export_use,
                valid_values=allowed_export_use)

        allowed_include_in_export = self.get_allowed_include_in_export_values()
        if include_in_export not in allowed_include_in_export:
            # Check if value contains comma and validate each item separately
            # Older iDRACs used to include comma separated option in
            # AllowableValues but got removed in newer versions violating
            # AllowableValues validation logic.
            include_in_export_rev =\
                mgr_maps.INCLUDE_EXPORT_VALUE_MAP_REV.get(include_in_export)
            all_items_valid = False
            if include_in_export_rev is not None:
                items = include_in_export_rev.split(',')
                all_items_valid = True
                for item in items:
                    if (mgr_maps.INCLUDE_EXPORT_VALUE_MAP[item]
                            not in allowed_include_in_export):
                        all_items_valid = False
                        break
            if not all_items_valid:
                raise sushy.exceptions.InvalidParameterValueError(
                    parameter='include_in_export', value=include_in_export,
                    valid_values=allowed_include_in_export)

        target = mgr_maps.EXPORT_CONFIG_VALUE_MAP_REV[target]
        export_use = mgr_maps.EXPORT_USE_VALUE_MAP_REV[export_use]
        include_in_export =\
            mgr_maps.INCLUDE_EXPORT_VALUE_MAP_REV[include_in_export]

        action_data = {
            'ShareParameters': {
                'Target': target
            },
            'ExportFormat': "JSON",
            'ExportUse': export_use,
            'IncludeInExport': include_in_export
        }

        try:
            response = asynchronous.http_call(
                self._conn,
                'post',
                self.export_system_configuration_uri,
                data=action_data)

            LOG.info("Successfully exported system configuration "
                     "for %(target)s", {'target': target})

            return response

        except (sushy.exceptions.ExtensionError,
                sushy.exceptions.InvalidParameterValueError) as exc:
            LOG.error('Dell OEM export system configuration failed : %s', exc)
            raise

    def export_system_configuration(self):
        """Export system configuration.

        Exports ALL targets for cloning and includes password hashes and
        read-only attributes.

        :returns: Response object containing configuration details.
        :raises: InvalidParameterValueError on invalid target.
        :raises: ExtensionError on failure to perform requested
            operation
        """
        include_in_export = mgr_cons.INCLUDE_EXPORT_READ_ONLY_PASSWORD_HASHES
        return self._export_system_configuration(
            mgr_cons.EXPORT_TARGET_ALL,
            export_use=mgr_cons.EXPORT_USE_CLONE,
            include_in_export=include_in_export)

    def get_pxe_port_macs_bios(self, ethernet_interfaces_mac):
        """Get a list of pxe port MAC addresses for BIOS.

        :param ethernet_interfaces_mac: Dictionary of ethernet interfaces.
        :returns: List of pxe port MAC addresses.
        :raises: ExtensionError on failure to perform requested operation.
        """
        pxe_port_macs = []
        # Get NIC configuration
        nic_settings = self._export_system_configuration(
            target=mgr_cons.EXPORT_TARGET_NIC)

        if nic_settings.status_code != _RESPONSE_OK_CODE:
            error = (('An error occurred when attempting to export '
                     'the system configuration. Status code: %(code), '
                      'Error details: %(err)'),
                     {'code': nic_settings.status_code,
                      'err': nic_settings.__dict__})
            LOG.error(error)
            raise sushy.exceptions.ExtensionError(error=error)
        # Parse the exported system configuration for the NIC
        # ports that are set to PXE boot
        json_data = nic_settings.json()
        if _SYSTEM_CONFIG_TAG in json_data.keys():
            for root in json_data[_SYSTEM_CONFIG_TAG]['Components']:
                nic_id = root['FQDD']
                for child in root['Attributes']:
                    if child.get('Name') == "LegacyBootProto":
                        if child['Value'] == "PXE":
                            mac_address = ethernet_interfaces_mac[nic_id]
                            pxe_port_macs.append(mac_address)
            return pxe_port_macs

        else:
            error = (('Failed to get system configuration from response'))
            LOG.error(error)
            raise sushy.exceptions.ExtensionError(error=error)

    def get_allowed_import_shutdown_type_values(self):
        """Get the allowed shutdown types of import system configuration.

        :returns: A set of allowed shutdown type values.
        """
        import_action = self._actions.import_system_configuration
        allowed_values = import_action.allowed_shutdown_type_values

        if not allowed_values:
            LOG.warning('Could not figure out the allowed values for the '
                        'shutdown type of import system configuration at %s',
                        self.path)
            return set(mgr_maps.IMPORT_SHUTDOWN_VALUE_MAP_REV)

        return set([mgr_maps.IMPORT_SHUTDOWN_VALUE_MAP[value] for value in
                    set(mgr_maps.IMPORT_SHUTDOWN_VALUE_MAP).
                    intersection(allowed_values)])

    def import_system_configuration(self, import_buffer):
        """Imports system configuration.

        Caller needs to handle system reboot separately.

        :param import_buffer: Configuration data to be imported.
        :returns: Task monitor instance to watch for task completion
        """
        action_data = dict(self.ACTION_DATA, ImportBuffer=import_buffer)
        # Caller needs to handle system reboot separately to preserve
        # one-time boot settings.
        shutdown_type = mgr_cons.IMPORT_SHUTDOWN_NO_REBOOT

        allowed_shutdown_types = self.get_allowed_import_shutdown_type_values()
        if shutdown_type not in allowed_shutdown_types:
            raise sushy.exceptions.InvalidParameterValueError(
                parameter='shutdown_type', value=shutdown_type,
                valid_values=allowed_shutdown_types)

        action_data['ShutdownType'] =\
            mgr_maps.IMPORT_SHUTDOWN_VALUE_MAP_REV[shutdown_type]

        response = self._conn.post(self.import_system_configuration_uri,
                                   data=action_data)

        return TaskMonitor.from_response(
            self._conn, response, self.import_system_configuration_uri)

    def reset_idrac(self, wait=True, ready_wait_time=60):
        """Reset the iDRAC and wait for it to become ready.

        :param wait: Whether to return immediately or wait for iDRAC to
            become operational.
        :param ready_wait_time: Amount of time in seconds to wait before
            starting to check on the iDRAC's status.
        """
        self.idrac_card_service.reset_idrac()
        if not wait:
            return
        host = urlparse(self._conn._url).netloc
        LOG.debug("iDRAC %(host)s was reset, "
                  "waiting for return to operational state", {'host': host})
        self._wait_for_idrac(host, ready_wait_time)
        self._wait_until_idrac_is_ready(host, self._IDRAC_IS_READY_RETRIES,
                                        self._IDRAC_IS_READY_RETRY_DELAY_SEC)

    def _wait_for_idrac_state(self, host, alive=True, ping_count=3,
                              retries=24):
        """Wait for iDRAC to become pingable or not pingable.

        :param host: Hostname or IP of the iDRAC interface.
        :param alive: True for pingable state and False for not pingable
            state.
        :param ping_count: Number of consecutive ping results, per
            'alive', for success.
        :param retries: Number of ping retries.
        :returns: True on reaching specified host ping state; otherwise,
            False.
        """
        if alive:
            ping_type = "pingable"
        else:
            ping_type = "not pingable"
        LOG.debug("Waiting for iDRAC %(host)s to become %(ping_type)s",
                  {'host': host, 'ping_type': ping_type})
        response_count = 0
        while retries > 0:
            response = self._ping_host(host)
            retries -= 1
            if response == alive:
                response_count += 1
                LOG.debug("iDRAC %(host)s is %(ping_type)s, "
                          "count=%(response_count)s",
                          {'host': host, 'ping_type': ping_type,
                           'response_count': response_count})
                if response_count == ping_count:
                    LOG.debug("Reached specified %(alive)s count for iDRAC "
                              "%(host)s", {'alive': alive, 'host': host})
                    return True
            else:
                response_count = 0
                if alive:
                    LOG.debug("iDRAC %(host)s is still not pingable",
                              {'host': host})
                else:
                    LOG.debug("iDRAC %(host)s is still pingable",
                              {'host': host})
            time.sleep(10)
        return False

    def _wait_for_idrac(self, host, post_pingable_wait_time):
        """Wait for iDRAC to transition from unpingable to pingable.

        :param host: Hostname or IP of the iDRAC interface.
        :param post_pingable_wait_time: Amount of time in seconds to
            wait after the host becomes pingable.
        :raises: ExtensionError on failure to perform requested
            operation.
        """
        state_reached = self._wait_for_idrac_state(
            host, alive=False, ping_count=2, retries=24)
        if not state_reached:
            error_msg = ("Timed out waiting iDRAC %(host)s to become not "
                         "pingable", {'host': host})
            LOG.error(error_msg)
            raise sushy.exceptions.ExtensionError(error=error_msg)
        LOG.debug("iDRAC %(host)s has become not pingable", {'host': host})
        state_reached = self._wait_for_idrac_state(host, alive=True,
                                                   ping_count=3, retries=24)
        if not state_reached:
            error_msg = ("Timed out waiting iDRAC %(host)s to become pingable",
                         {'host': host})
            LOG.error(error_msg)
            raise sushy.exceptions.ExtensionError(error=error_msg)
        LOG.debug("iDRAC %(host)s has become pingable", {'host': host})
        time.sleep(post_pingable_wait_time)

    def _wait_until_idrac_is_ready(self, host, retries, retry_delay):
        """Wait until the iDRAC is in a ready state.

        :param host: Hostname or IP of the iDRAC interface.
        :param retries: The number of times to check if the iDRAC is
            ready.
        :param retry_delay: The number of seconds to wait between
            retries.
        :raises: ExtensionError on failure to perform requested
            operation.
        """

        while retries > 0:
            LOG.debug("Checking to see if iDRAC %(host)s is ready",
                      {'host': host})
            if self.lifecycle_service.is_idrac_ready():
                LOG.debug("iDRAC %(host)s is ready", {'host': host})
                return
            LOG.debug("iDRAC %(host)s is not ready", {'host': host})
            retries -= 1
            if retries > 0:
                time.sleep(retry_delay)
        if retries == 0:
            error_msg = ("Timed out waiting iDRAC %(host)s to become "
                         "ready after reset", {'host': host})
            LOG.error(error_msg)
            raise sushy.exceptions.ExtensionError(error=error_msg)

    def _ping_host(self, host):
        """Ping the hostname or IP of a host.

        :param host: Hostname or IP.
        :returns: True if host is alive; otherwise, False.
        """
        response = subprocess.call(["ping", "-c", "1", host])
        return response == 0


def get_extension(*args, **kwargs):
    return DellManagerExtension



# Copyright (c) 2020 Dell Inc. or its subsidiaries.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslotest.base import BaseTestCase
import sushy

from sushy_oem_idrac.asynchronous import http_call


class AsychronousTestCase(BaseTestCase):

    def setUp(self):
        super(AsychronousTestCase, self).setUp()
        self.conn = mock.Mock()

    def test_http_call_post_accepted(self):
        mock_post_response = self.conn.post.return_value
        mock_post_response.status_code = 202
        mock_post_response.headers.get.return_value = '1'

        mock_get_202_response = mock.Mock()
        mock_get_202_response.status_code = 202
        mock_get_202_response.headers.get.return_value = '1'

        mock_get_200_response = mock.Mock()
        mock_get_200_response.status_code = 200

        self.conn.get.side_effect = [
            mock_get_202_response, mock_get_200_response]

        resp = http_call(self.conn, 'POST')
        self.assertIs(resp, mock_get_200_response)

    def test_http_call_post_accepted_no_location(self):
        mock_response = self.conn.post.return_value
        mock_response.status_code = 202
        mock_response.headers.get.return_value = None

        self.assertRaises(sushy.exceptions.ExtensionError,
                          http_call, self.conn, 'POST')



# Copyright (c) 2021 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Job state constants
JOB_STATE_COMPLETED = "Completed"
"""A job is in completed state"""

JOB_STATE_COMPLETED_ERRORS = "Completed with errors"
"""A job is in completed state with errors"""

JOB_STATE_DOWNLOADED = "Downloaded"
"""A job is in downloaded state"""

JOB_STATE_DOWNLOADING = "Downloading"
"""A job is in downloading state"""

JOB_STATE_FAILED = "Failed"
"""A job is in failed state"""

JOB_STATE_NEW = "New"
"""A job is in newly created state"""

JOB_STATE_PAUSED = "Paused"
"""A job is in paused state"""

JOB_STATE_PENDING_ACTIVATION = "Pending activation"
"""A job is in pending activation state"""

JOB_STATE_READY_EXECUTION = "Ready for execution"
"""A job is in ready for execution state"""

JOB_STATE_REBOOT_COMPLETED = "Reboot completed"
"""A job is in reboot completed state"""

JOB_STATE_REBOOT_FAILED = "Reboot failed"
"""A job is in reboot failed state"""

JOB_STATE_REBOOT_PENDING = "Reboot pending"
"""A job is in pending state for reboot"""

JOB_STATE_RUNNING = "Running"
"""A job is in running state"""

JOB_STATE_SCHEDULED = "Scheduled"
"""A job is in scheduled state"""

JOB_STATE_SCHEDULING = "Scheduling"
"""A job is in scheduling state"""

JOB_STATE_UNKNOWN = "Unknown"
"""A job is in unknown state"""

JOB_STATE_WAITING = "Waiting"
"""A job is in waiting state"""

# Job type constants
JOB_TYPE_BIOS_CONF = "BIOS configuration"
"""A BIOS configuration job"""

JOB_TYPE_EXPORT_CONF = "Export configuration"
"""A server configuration profile export job"""

JOB_TYPE_FC_CONF = "Fibre Channel configuration"
"""A Fibre Channel configuration job"""

JOB_TYPE_FACTORY_CONF_EXPORT = "Factory configuration export"
"""A factory configuration export job"""

JOB_TYPE_FIRMWARE_ROLLBACK = "Firmware rollback"
"""A firmware rollback job"""

JOB_TYPE_FIRMWARE_UPDATE = "Firmware update"
"""A firmware update job"""

JOB_TYPE_HW_INVENTORY_EXPORT = "Hardware inventory export"
"""A hardware inventory export job"""

JOB_TYPE_IMPORT_CONF = "Import configuration"
"""A server configuration profile import job"""

JOB_TYPE_INBAND_BIOS_CONF = "Inband BIOS configuration"
"""An inband BIOS configuration job"""

JOB_TYPE_LC_CONF = "LC configuration"
"""A lifecycle controller attribute configuration job"""

JOB_TYPE_LC_EXPORT = "LC export"
"""A lifecycle controller export job"""

JOB_TYPE_LC_LOG_EXPORT = "LC log export"
"""A lifecycle controller log export job"""

JOB_TYPE_LICENSE_EXPORT = "License export"
"""A license export job"""

JOB_TYPE_LICENSE_IMPORT = "License import"
"""A license import job"""

JOB_TYPE_MSG_REG_EXPORT = "Message registry export"
"""Export message registry report job"""

JOB_TYPE_NIC_CONF = "NIC configuration"
"""A NIC configuration job"""

JOB_TYPE_OS_DEPLOY = "OS deploy"
"""Operating System deploy job"""

JOB_TYPE_RAID_CONF = "RAID configuration"
"""A RAID configuration job"""

JOB_TYPE_RT_NO_REBOOT_CONF = "Real-time no reboot configuration"
"""A real time configuration job without reboot"""

JOB_TYPE_REBOOT_FORCE = "Reboot force"
"""A reboot job with forced shutdown"""

JOB_TYPE_REBOOT_NO_FORCE = "Reboot no force"
"""A graceful reboot job without forced shutdown"""

JOB_TYPE_REBOOT_POWER_CYCLE = "Reboot power cycle"
"""A power cycle job"""

JOB_TYPE_REMOTE_DIAG = "Remote diagnostics"
"""A remote diagnostics job"""

JOB_TYPE_REPO_UPDATE = "Repository update"
"""An update job from a repository"""

JOB_TYPE_SA_COL_EXP_HEALTH_DATA = "SA collect and export health data"
"""Support Assist collect and export health data job"""

JOB_TYPE_SA_COL_HEALTH_DATA = "SA collect health data"
"""Support Assist collect health data job"""

JOB_TYPE_SA_EXP_HEALTH_DATA = "SA export health data"
"""Support Assist export health data job"""

JOB_TYPE_SA_ISM = "SA expose iSM"
"""Support Assist expose iDRAC Service Module installer package to host job"""

JOB_TYPE_SA_REG = "SA registration"
"""Support Assist register iDRAC to Dell backend server job"""

JOB_TYPE_SEKM_REKEY = "SEKM rekey"
"""A Secure Enterprise Key Manager rekey job"""

JOB_TYPE_SEKM_STATUS_SET = "SEKM status set"
"""A Secure Enterprise Key Manager status set job"""

JOB_TYPE_SHUTDOWN = "Shutdown"
"""A shutdown job"""

JOB_TYPE_SYS_ERASE = "System erase"
"""A selective system erase job"""

JOB_TYPE_SYS_INFO_CONF = "System info configuration"
"""A system info configuration job"""

JOB_TYPE_THERMAL_HIST_EXP = "Thermal history export"
"""A thermal history export job"""

JOB_TYPE_UNKNOWN = "Unknown"
"""An unknown job"""

JOB_TYPE_IDRAC_CONF = "iDRAC configuration"
"""An iDRAC configuration job"""



# Copyright (c) 2021 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from sushy import utils

from sushy_oem_idrac.resources.taskservice import constants as ts_cons

JOB_STATE_VALUE_MAP = {
    "New": ts_cons.JOB_STATE_NEW,
    "Scheduled": ts_cons.JOB_STATE_SCHEDULED,
    "Running": ts_cons.JOB_STATE_RUNNING,
    "Completed": ts_cons.JOB_STATE_COMPLETED,
    "Downloading": ts_cons.JOB_STATE_DOWNLOADING,
    "Downloaded": ts_cons.JOB_STATE_DOWNLOADED,
    "Scheduling": ts_cons.JOB_STATE_SCHEDULING,
    "ReadyForExecution": ts_cons.JOB_STATE_READY_EXECUTION,
    "Waiting": ts_cons.JOB_STATE_WAITING,
    "Paused": ts_cons.JOB_STATE_PAUSED,
    "Failed": ts_cons.JOB_STATE_FAILED,
    "CompletedWithErrors": ts_cons.JOB_STATE_COMPLETED_ERRORS,
    "RebootPending": ts_cons.JOB_STATE_REBOOT_PENDING,
    "RebootFailed": ts_cons.JOB_STATE_REBOOT_FAILED,
    "RebootCompleted": ts_cons.JOB_STATE_REBOOT_COMPLETED,
    "PendingActivation": ts_cons.JOB_STATE_PENDING_ACTIVATION,
    "Unknown": ts_cons.JOB_STATE_UNKNOWN
}

JOB_STATE_VALUE_MAP_REV = utils.revert_dictionary(JOB_STATE_VALUE_MAP)

JOB_TYPE_VALUE_MAP = {
    "FirmwareUpdate": ts_cons.JOB_TYPE_FIRMWARE_UPDATE,
    "FirmwareRollback": ts_cons.JOB_TYPE_FIRMWARE_ROLLBACK,
    "RepositoryUpdate": ts_cons.JOB_TYPE_REPO_UPDATE,
    "RebootPowerCycle": ts_cons.JOB_TYPE_REBOOT_POWER_CYCLE,
    "RebootForce": ts_cons.JOB_TYPE_REBOOT_FORCE,
    "RebootNoForce": ts_cons.JOB_TYPE_REBOOT_NO_FORCE,
    "Shutdown": ts_cons.JOB_TYPE_SHUTDOWN,
    "RAIDConfiguration": ts_cons.JOB_TYPE_RAID_CONF,
    "BIOSConfiguration": ts_cons.JOB_TYPE_BIOS_CONF,
    "NICConfiguration": ts_cons.JOB_TYPE_NIC_CONF,
    "FCConfiguration": ts_cons.JOB_TYPE_FC_CONF,
    "iDRACConfiguration": ts_cons.JOB_TYPE_IDRAC_CONF,
    "SystemInfoConfiguration": ts_cons.JOB_TYPE_SYS_INFO_CONF,
    "InbandBIOSConfiguration": ts_cons.JOB_TYPE_INBAND_BIOS_CONF,
    "ExportConfiguration": ts_cons.JOB_TYPE_EXPORT_CONF,
    "ImportConfiguration": ts_cons.JOB_TYPE_IMPORT_CONF,
    "RemoteDiagnostics": ts_cons.JOB_TYPE_REMOTE_DIAG,
    "RealTimeNoRebootConfiguration": ts_cons.JOB_TYPE_RT_NO_REBOOT_CONF,
    "LCLogExport": ts_cons.JOB_TYPE_LC_LOG_EXPORT,
    "HardwareInventoryExport": ts_cons.JOB_TYPE_HW_INVENTORY_EXPORT,
    "FactoryConfigurationExport": ts_cons.JOB_TYPE_FACTORY_CONF_EXPORT,
    "LicenseImport": ts_cons.JOB_TYPE_LICENSE_IMPORT,
    "LicenseExport": ts_cons.JOB_TYPE_LICENSE_EXPORT,
    "ThermalHistoryExport": ts_cons.JOB_TYPE_THERMAL_HIST_EXP,
    "LCConfig": ts_cons.JOB_TYPE_LC_CONF,
    "LCExport": ts_cons.JOB_TYPE_LC_EXPORT,
    "SACollectHealthData": ts_cons.JOB_TYPE_SA_COL_HEALTH_DATA,
    "SAExportHealthData": ts_cons.JOB_TYPE_SA_EXP_HEALTH_DATA,
    "SACollectExportHealthData": ts_cons.JOB_TYPE_SA_COL_EXP_HEALTH_DATA,
    "SAExposeISM": ts_cons.JOB_TYPE_SA_ISM,
    "SARegistration": ts_cons.JOB_TYPE_SA_REG,
    "SystemErase": ts_cons.JOB_TYPE_SYS_ERASE,
    "MessageRegistryExport": ts_cons.JOB_TYPE_MSG_REG_EXPORT,
    "OSDeploy": ts_cons.JOB_TYPE_OS_DEPLOY,
    "SEKMRekey": ts_cons.JOB_TYPE_SEKM_REKEY,
    "SEKMStatusSet": ts_cons.JOB_TYPE_SEKM_STATUS_SET,
    "Unknown": ts_cons.JOB_TYPE_UNKNOWN
}

JOB_TYPE_VALUE_MAP_REV = utils.revert_dictionary(JOB_TYPE_VALUE_MAP)
